/*
 *
 * a)
 *
 * 1: Om två trådar kör raden:
 *
 * int my_event = ani->event++;
 *
 * samtidigt så kan båda få samma värde i my_event, eftersom ++-operatorn måste
 * implementeras som:
 *
 * int tmp = ani->event;
 * tmp++;
 * ani->event = tmp;
 * int my_event = tmp;
 *
 * Om två trådar läser 'ani->event' samtidigt kommer båda få samma
 * värde i 'tmp'. Detta kommer också göra att 'event' bara ökas en
 * gång i stället för två.
 *
 * 2: Liknande situation i "animations_add":
 *
 * Om två trådar kör raden:
 *
 * ani->events[ani->event_count++] = event;
 *
 * Samtidigt så kommer de av samma anledning som ovan spara 'event' på
 * samma plats i 'events' arrayen. Därmed blir en av dem överskriven.
 *
 * 3: Koden i 'animations_run_step' väntar bara på att variabeln
 * 'next_event' har ökats så att den är större än eller lika med
 * 'event_count'.
 *
 * Eftersom koden i 'animations_worker' ökar på variabeln *innan* den
 * faktiskt kör händelsen, så kommer 'animations_run_step' att bara
 * vänta på att någon tråd *kommer* att utföra händelsen, inte att den
 * har körts klart. Det gör att 'animations_run_step' kan returnera
 * för tidigt.
 *
 * 4: Detta liknar problemet ovan. Problemet här är att
 * 'animations_destroy' inte väntar på att trådarna faktiskt har
 * avslutats. När 'animations_destroy' signalerar till trådarna att
 * avsluta så kommer den därför sedan direkt att avallokera minne.
 * Därmed, om trådarnana som kör 'animations_worker' startar efter att
 * minnet har avallokerats så kan de krascha när de avrefererar
 * 'ani'-pekaren.
 *
 * b) Busy-wait fanns i följande loopar:
 *
 * I 'animation_worker':
 *
 * while (!ani->start_work)
 *     ;
 *
 * I 'animations_run_step':
 *
 * while (ani->next_event < ani->event_count)
 *     ;
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"

struct event;
struct animations;

// Typ för att lagra funktionspekare. Lagrar pekare till funktioner
// som ser ut som följer:
//   void handler(struct animations *, struct event *)
typedef void (*event_handler_fn)(struct animations *, struct event *);


/**
 * Datastruktur som representerar något som ska hända under en
 * animation. Lagrar en pekare till en funktion som implementerar
 * händelsen tillsammans med ett heltal som lagrar godtycklig data.
 */
struct event {
    // Funktion som ska anropas.
    event_handler_fn handler;

    // Data som funktionen kan använda.
    int data;
};


/**
 * Datastruktur som representerar pågående animationer i form av en
 * samling händelser (event). Alla händelser kan sedan köras med
 * funktionen 'animations_run_step'.
 *
 * Du kan anta att innehållet i datastrukturen bara används av
 * funktionerna som börjar med 'animations_' här nedanför. Det är
 * alltså okej att ersätta variabler här med synkroniseringsprimitiver
 * vid behov.
 */
struct animations {
    // Array av händelser i animationen.
    struct event *events;
    // Antal platser i 'events' som används just nu.
    int event_count;
    // Totalt antal platser som finns i 'events'.
    int event_max;

    // Antal trådar som är redo att hantera de händelser som
    // finns i 'events'.
    int thread_count;
    // Vilken händelse är nästa som trådarna ska behandla?
    int next_event;
    // Används för att meddela trådarna att det är dags att arbeta.
    // bool start_work; // ersatt av semafor.
    // Används för att informera trådarna om att det är dags att
    // avslutas.
    bool exit;

    // Lås för 'event_count'.
    struct lock event_count_lock;
    // Lås för 'next_event'.
    struct lock next_event_lock;
    // Semafor för att väcka trådarna.
    struct semaphore start_work;
    // Semafor för att signalera att trådarna är klara.
    struct semaphore done_work;
};


// Funktion som körs av trådarna som startas av 'animations_create'.
// Trådarna väntar på att få information om att det är dags att arbeta
static void animations_worker(struct animations *ani) {
    bool should_exit = false;
    do {
        // Är det dags att göra något?
        sema_down(&ani->start_work);

        // Ska vi stänga av oss?
        should_exit = ani->exit;

        while (!should_exit) {
            // Finns det något att göra?
            bool done = false;

            lock_acquire(&ani->next_event_lock);
            lock_acquire(&ani->event_count_lock);
            if (ani->next_event >= ani->event_count) {
                done = true;
            }
            // Plocka en händelse att behandla.
            int my_event = ani->next_event++;
            lock_release(&ani->next_event_lock);
            lock_release(&ani->event_count_lock);

            if (done)
                break;

            // Kör funktionen i händelsen.
            struct event *e = &ani->events[my_event];
            (*e->handler)(ani, e);
        }

        // Meddela att vi är klara.
        sema_up(&ani->done_work);

    } while (!should_exit);
}


// Skapa ett animations-objekt med plats för maximalt 'event_max'
// händelser. Startar också 'thread_count' trådar för att hantera
// händelserna parallellt när 'animations_run_step' anropas.
struct animations *animations_create(int events_max, int thread_count) {
    struct animations *ani = malloc(sizeof(struct animations));
    ani->events = malloc(sizeof(struct event) * events_max);
    ani->event_count = 0;
    ani->event_max = events_max;
    ani->thread_count = thread_count;
    ani->next_event = 0;
    // ani->start_work = false;
    ani->exit = false;

    lock_init(&ani->event_count_lock);
    lock_init(&ani->next_event_lock);
    sema_init(&ani->start_work, 0);
    sema_init(&ani->done_work, 0);

    // Starta trådar så att de är redo.
    for (int i = 0; i < thread_count; i++)
        thread_new(&animations_worker, ani);

    return ani;
}

// Lägg till en händelse till ett animationsobjekt. Detta ska kunna
// göras när som helst så länge animationsobjektet inte förstörs.
bool animations_add(struct animations *ani, struct event event) {
    lock_acquire(&ani->event_count_lock);
    if (ani->event_count >= ani->event_max) {
        lock_release(&ani->event_count_lock);
        return false;
    }

    ani->events[ani->event_count++] = event;
    lock_release(&ani->event_count_lock);

    return true;
}

// Kör alla händelser som finns i animationsobjektet. Du kan anta att
// det maximalt är en tråd som anropar 'animations_run_step' för ett
// specifikt anrop samtidigt. Däremot ska det gå att anropa
// 'animations_run_step' för olika animationsobjekt samtidigt.
void animations_run_step(struct animations *ani) {
    ani->next_event = 0;

    for (int i = 0; i < ani->thread_count; i++)
        sema_up(&ani->start_work);

    for (int i = 0; i < ani->thread_count; i++)
        sema_down(&ani->done_work);
}

// Förstör ett animationsobjekt. Avslutar de trådar som skapades
// i 'animations_create' och avallokerar minnet för datastrukturen.
// Du kan anta att ingen annan tråd använder animationsobjektet
// när 'animations_destroy' anropas.
void animations_destroy(struct animations *ani) {
    // Be trådarna att avsluta sig själva, och väck dem sedan
    // genom att sätta 'start_work' till 'true'.
    ani->exit = true;

    for (int i = 0; i < ani->thread_count; i++)
        sema_up(&ani->start_work);

    for (int i = 0; i < ani->thread_count; i++)
        sema_down(&ani->done_work);

    // Frigör minnet.
    free(ani->events);
    free(ani);
}

/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */

// Lås för stdout.
struct lock stdout_lock;

// Händelse som skriver ut ett tecken till stdout.
void print_handler(struct animations *a, struct event *e) {
    (void)a;
    lock_acquire(&stdout_lock);
    printf("%c", (char)e->data);
    lock_release(&stdout_lock);
}

// Skapa en händelse som anropar 'print_handler' ovan och skriver ut
// ett tecken till skärmen.
void add_print_handler(struct animations *to, char to_print) {
    struct event event = {
        print_handler,
        (int)to_print
    };
    animations_add(to, event);
}

// Händelse som lägger till en ny händelse varje steg.
void create_handler(struct animations *a, struct event *e) {
    add_print_handler(a, (char)(e->data++));
}

// Skapa en händelse som lägger till nya händelser varje gång.
void add_create_handler(struct animations *to, char start) {
    struct event event = {
        create_handler,
        (int)start
    };
    animations_add(to, event);
}

// Main, använder 'animations' för att köra händelser i 10 steg.
int main() {
    const char *divider = "--------------------";
    lock_init(&stdout_lock);

    // Skapa en animation.
    struct animations *ani = animations_create(40, 4);

    // Lägg till en händelse som lägger till nya utskrifter av tecken varje iteration.
    add_create_handler(ani, 'A');

    // Lägg till några vanliga utskrifter till att börja med.
    for (int i = 0; i < 10; i++) {
        add_print_handler(ani, 'a' + i);
    }

    // Kör animationen i 10 steg.
    printf("%s\n", divider);
    for (int i = 0; i < 10; i++) {
        animations_run_step(ani);
        printf("\n%s\n", divider);
    }

    // Städa upp.
    animations_destroy(ani);

    return 0;
}
