/**
 *
 * a)
 *
 * 1: Om två trådar kör någon av raderna "f->threads_done++" samtidigt
 * så finns det en risk att något inkrement "glöms bort". Detta för
 * att inkrement-operatorn implementeras som följer:
 *
 * int tmp = f->threads_done;
 * tmp++;
 * f->threads_done = tmp;
 *
 * Om två trådar kör raderna då 'threads_done' exempelvis är 0, så
 * kommer båda få 'tmp = 0', öka 'tmp' till 1, och till sist sätta
 * 'threads_done' till 1. Detta trots att 'threads_done' borde ha
 * blivit 2.
 *
 * 2: Om två trådar kör 'multi_post_result' samtidigt så kan båda
 * kontrollera 'has_result' samtidigt, och båda konstaterar att de
 * borde spara resultatet. Det gör att båda skriver till 'result', men
 * det ena resultatet blir överskrivet.
 *
 * b) Följande busy-wait finns i koden i 'multi_wait':
 *
 * while (f->threads_done < threads)
 *     ;
 *
 * e)
 *
 * Alla delade variabler är skyddade. Notera att 'has_result' och
 * 'result' inte behöver skyddas i 'multi_wait' eftersom vi då vet att
 * ingen annan tråd kan få för sig att skriva till dem samtidigt.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"

/**
 * Datastruktur för att hantera multipla trådar som ska producera ett
 * resultat i slutändan.
 *
 * Du kan anta att innehållet i datastrukturen bara används av
 * funktionerna som börjar med 'multi_' här nedanför. Det är alltså
 * okej att ersätta variabler här med synkroniseringsprimitiver vid
 * behov.
 */
struct multiresult {
    // Slutgiltigt resultat. Endast giltigt om 'has_result' är sann.
    int64_t result;
    // Har vi ett resultat i 'result'?
    bool has_result;
    // Lås för 'result' och 'has_result'.
    struct lock result_lock;
    // Hur många trådar är klara? (= antal trådar som har anropat
    // antingen 'multi_post_nothing' eller 'multi_post_result')
    struct semaphore threads_done;
};

// Initiera en multiresult-struktur.
void multi_init(struct multiresult *f) {
    f->result = 0;
    f->has_result = false;
    lock_init(&f->result_lock);
    sema_init(&f->threads_done, 0);
}

// Kontrollera om objektet innehåller ett resultat eller inte.
bool multi_has_result(struct multiresult *f) {
    lock_acquire(&f->result_lock);
    bool r = f->has_result;
    lock_release(&f->result_lock);
    return r;
}

// Signalera att en tråd är klar, och att den inte producerade ett
// resultat.
void multi_post_nothing(struct multiresult *f) {
    sema_up(&f->threads_done);
}

// Signalera att en tråd är klar, och att den producerade resultatet
// 'value'. Om flera trådar anropar funktionen så ska bara det första
// resultatet lagras.
void multi_post_result(struct multiresult *f, int64_t value) {
    lock_acquire(&f->result_lock);
    if (!f->has_result) {
        f->result = value;
        f->has_result = true;
    }
    lock_release(&f->result_lock);
    sema_up(&f->threads_done);
}

// Vänta på att 'threads' trådar har signalerat att de är klara (med
// 'multi_post_nothing' eller 'multi_post_result'). Om åtminstone en
// tråd producerade ett resultat ska funktionen returnera 'true' och
// sätta '*out' till resultatet. Om ingen tråd producerade ett
// resultat ska funktionen returnera 'false' och '*out' ska lämnas
// orörd. Du kan anta att 'multi_wait' anropas maximalt en gång för
// varje 'multiresult'-objekt.
bool multi_wait(struct multiresult *f, int64_t *out, int threads) {
    for (int i = 0; i < threads; i++)
        sema_down(&f->threads_done);

    if (f->has_result) {
        *out = f->result;
        return true;
    } else {
        return false;
    }
}

/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */

// Delad data mellan trådarna.
struct worker_data {
    // Synkronisering.
    struct multiresult multiresult;

    // Tal som ska kontrolleras.
    int64_t to_check;
    // Steglängd (= antal trådar).
    int64_t step;
    // Heltal som används för att se till att trådarna får olika start-tal.
    int64_t start;
};

// Funktion som körs i varje tråd.
void worker(struct worker_data *d) {
    // Hämta ett tal som är unikt för oss.
    int64_t divisor = atomic_add(&d->start, 1);

    // Itererera genom alla tal, avbryt om någon annan tråd har hittat ett resultat.
    while (divisor < d->to_check && !multi_has_result(&d->multiresult)) {

        // Har vi hittat en delare?
        if (d->to_check % divisor == 0) {
            multi_post_result(&d->multiresult, divisor);
            return;
        }

        // Stega framåt.
        divisor += d->step;
    }

    multi_post_nothing(&d->multiresult);
}

// Kontrollera om talet 'to_check' är ett primtal med hjälp av 'num_threads' trådar.
void check_prime(int64_t to_check, int num_threads) {
    struct worker_data data;
    multi_init(&data.multiresult);
    data.to_check = to_check;
    data.step = num_threads;
    data.start = 2;

    for (int i = 0; i < num_threads; i++)
        thread_new(&worker, &data);

    int64_t result = 0;
    if (multi_wait(&data.multiresult, &result, num_threads)) {
        printf("%ld is not prime, one divisor is %ld\n", to_check, result);
    } else {
        printf("%ld is prime!\n", to_check);
    }
}

// Main, testar 'check_prime' med ett par exempel.
int main(void) {
    check_prime(46061 * 46073, 8);
    check_prime(2147400001L * 21400003L, 8);
    check_prime(31400023L, 8);

    return 0;
}
